/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2015 ResiliNets, ITTC, University of Kansas
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Authors: Truc Anh N. Nguyen <annguyen@ittc.ku.edu>
 *
 * James P.G. Sterbenz <jpgs@ittc.ku.edu>, director
 * ResiliNets Research Group  http://wiki.ittc.ku.edu/resilinets
 * Information and Telecommunication Technology Center (ITTC)
 * and Department of Electrical Engineering and Computer Science
 * The University of Kansas Lawrence, KS USA.
 */

#define NS_LOG_APPEND_CONTEXT \
  if (m_node) { std::clog << Simulator::Now ().GetSeconds () << " [node " << m_node->GetId () << "] "; }

#include "tcp-sack.h"
#include "ns3/log.h"
#include "ns3/trace-source-accessor.h"
#include "ns3/simulator.h"
#include "ns3/abort.h"
#include "ns3/node.h"
#include "ns3/packet.h"

NS_LOG_COMPONENT_DEFINE ("TcpSack");

namespace ns3 {

NS_OBJECT_ENSURE_REGISTERED (TcpSack);

TypeId
TcpSack::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::TcpSack")
    .SetParent<TcpSocketBase> ()
    .SetGroupName ("Internet")
    .AddConstructor<TcpSack> ()
    .AddAttribute ("ReTxThreshold", "Threshold for fast retransmit",
                   UintegerValue (3),
                   MakeUintegerAccessor (&TcpSack::m_retxThresh),
                   MakeUintegerChecker<uint32_t> ())
    .AddTraceSource ("CongestionWindow",
                     "The TCP connection's congestion window",
                     MakeTraceSourceAccessor (&TcpSack::m_cWnd),
                     "ns3::TracedValue::Uint32Callback")
    .AddTraceSource ("HighestRxtSequence",
                     "Highest sequence number ever retransmitted in socket's life time",
                     MakeTraceSourceAccessor (&TcpSack::m_highRxt),
                     "ns3::SequenceNumber32TracedValueCallback")
    .AddTraceSource ("RescueRxt",
                     "Highest sequence number has been optimistically retransmitted",
                     MakeTraceSourceAccessor (&TcpSack::m_rescueRxt),
                     "ns3::SequenceNumber32TracedValueCallback")
  ;
  return tid;
}

TcpSack::TcpSack (void)
  : m_retxThresh (3),
    m_inFastRec (false),
    m_highRxt (0),
    m_rescueRxt (0),
    m_sackDupAckCount (0)
{
  NS_LOG_FUNCTION (this);
}

TcpSack::TcpSack (const TcpSack& sock)
  : TcpSocketBase (sock),
    m_cWnd (sock.m_cWnd),
    m_ssThresh (sock.m_ssThresh),
    m_initialCWnd (sock.m_initialCWnd),
    m_retxThresh (sock.m_retxThresh),
    m_inFastRec (false),
    m_sackDupAckCount (sock.m_sackDupAckCount)
{
  NS_LOG_FUNCTION (this);
  NS_LOG_LOGIC ("Invoked the copy constructor");
}

TcpSack::~TcpSack (void)
{
}

int
TcpSack::Listen (void)
{
  NS_LOG_FUNCTION (this);
  InitializeCwnd ();
  return TcpSocketBase::Listen ();
}

int
TcpSack::Connect (const Address & address)
{
  NS_LOG_FUNCTION (this << address);
  InitializeCwnd ();
  return TcpSocketBase::Connect (address);
}

uint32_t
TcpSack::Window (void)
{
  NS_LOG_FUNCTION (this);
  // Limit the size of in-flight data by cwnd and receiver's rxwin
  return std::min (m_rWnd.Get (), m_cWnd.Get ());
}

Ptr<TcpSocketBase>
TcpSack::Fork (void)
{
  return CopyObject<TcpSack> (this);
}

void
TcpSack::ReceivedAck (Ptr<Packet> packet, const TcpHeader& tcpHeader)
{
  NS_LOG_FUNCTION (this << tcpHeader << m_rxSackList.empty ());
  bool isDupAck = false;
  if (!m_rxSackList.empty ())
    {
      isDupAck = m_scoreBoard.IsDupAck (m_rxSackList, m_txBuffer->HeadSequence (), m_highTxMark);
      NS_LOG_LOGIC ("Is it a dupack? " << isDupAck);
    }

  TcpSocketBase::ReceivedAck (packet, tcpHeader);

  if (isDupAck)
    {
      m_sackDupAckCount++;
      SackDupAck (tcpHeader, m_sackDupAckCount);
    }
}

void
TcpSack::NewAck (const SequenceNumber32& seq)
{
  NS_LOG_FUNCTION (this << seq);
  NS_LOG_LOGIC ("TcpSack receieved ACK for seq " << seq <<
                " cwnd " << m_cWnd <<
                " ssthresh " << m_ssThresh);

  // Reset dup ACK count to 0 upon the receipt of a cumulative ACK
  m_sackDupAckCount = 0;

  // Check for exit condition of fast recovery
  if (m_inFastRec && seq < m_recover)
    { // Partial ACK, partial window deflation (RFC2582 sec.3 bullet #5 paragraph 3)
      m_cWnd -= seq - m_txBuffer->HeadSequence ();
      m_cWnd += m_segmentSize;  // increase cwnd
      NS_LOG_INFO ("Partial ACK in fast recovery: cwnd set to " << m_cWnd);

      // Update m_nextTxSequence and send new data if allowed by window
      TcpSocketBase::NewAck (seq);

      // Assume the next seq is lost and retransmit
      uint32_t sz = DoRetransmitPacket ();

      // Update m_highRxt to prevent duplicate retransmission
      m_highRxt = std::max (m_txBuffer->HeadSequence () + sz, m_highRxt.Get ());

      // Start the "pipe" algorithm
      StartPipe ();
      return;
    }
  else if (m_inFastRec && seq >= m_recover)
    { // Full ACK (RFC2582 sec.3 bullet #5 paragraph 2, option 1)
      m_cWnd = std::min (m_ssThresh, BytesInFlight () + m_segmentSize);
      m_inFastRec = false;
      NS_LOG_INFO ("Received full ACK. Leaving fast recovery with cwnd set to " << m_cWnd);
    }

  // Increase of cwnd based on current phase (slow start or congestion avoidance)
  if (m_cWnd < m_ssThresh)
    { // Slow start mode, add one segSize to cWnd. Default m_ssThresh is 65535. (RFC2001, sec.1)
      m_cWnd += m_segmentSize;
      NS_LOG_INFO ("In SlowStart, updated to cwnd " << m_cWnd << " ssthresh " << m_ssThresh);
    }
  else
    { // Congestion avoidance mode, increase by (segSize*segSize)/cwnd. (RFC2581, sec.3.1)
      // To increase cwnd for one segSize per RTT, it should be (ackBytes*segSize)/cwnd
      double adder = static_cast<double> (m_segmentSize * m_segmentSize) / m_cWnd.Get ();
      adder = std::max (1.0, adder);
      m_cWnd += static_cast<uint32_t> (adder);
      NS_LOG_INFO ("In CongAvoid, updated to cwnd " << m_cWnd << " ssthresh " << m_ssThresh);
    }

  // Complete newAck processing
  TcpSocketBase::NewAck (seq);
}

void
TcpSack::DupAck (const TcpHeader& header, uint32_t count)
{
}

void
TcpSack::SackDupAck (const TcpHeader& header, uint32_t count)
{
  NS_LOG_FUNCTION (this << count);

  if ((count >= m_retxThresh && !m_inFastRec) || ((count < m_retxThresh) && m_scoreBoard.IsLost (m_txBuffer->HeadSequence (), m_retxThresh)))
    {
      if ((m_afterRTO && m_txBuffer->HeadSequence () >= m_recover) || !m_afterRTO)
        { // if after an RTO, check m_recover and high ACK

          // Re-set m_afterRTO if TRUE
          if (m_afterRTO)
            {
              m_afterRTO = false;
            }

          // Set the recovery point to the highest sequence transmitted
          m_recover = m_highTxMark;

          // Reduce the cwnd and ssthresh to 1/2 FlightSize
          m_ssThresh = std::max (2 * m_segmentSize, BytesInFlight () / 2);
          m_cWnd = m_ssThresh + 3 * m_segmentSize;

          // Enter the fast recovery mode
          m_inFastRec = true;
          NS_LOG_INFO ("Triple dupack. Set cwnd to " << m_cWnd << ", ssthresh to " << m_ssThresh << ", recovery point to " << m_recover);

          // Retransmit the first data segment presumed dropped
          uint32_t sz = DoRetransmitPacket ();

          // Update m_highRxt and m_rescueRxt to prevent duplicate retransmission
          m_highRxt = std::max (m_txBuffer->HeadSequence () + sz, m_highRxt.Get ());
          m_rescueRxt = std::max (m_txBuffer->HeadSequence () + sz, m_rescueRxt.Get ());

          // Start the "pipe" algorithm
          StartPipe ();
        }
    }
  else if (m_inFastRec)
    { // In fast recovery, inc cwnd for every additional dupack (RFC2581, sec.3.2)
      m_cWnd += m_segmentSize;
      NS_LOG_INFO ("Increased cwnd to " << m_cWnd);
      SendPendingData (m_connected);
    }
  else if (!m_inFastRec && m_limitedTx && m_txBuffer->SizeFromSequence (m_nextTxSequence) > 0)
    { // RFC3042 Limited transmit: Send a new packet for each duplicated ACK before fast retransmit
      NS_LOG_INFO ("Limited transmit");
      m_highRxt = m_txBuffer->HeadSequence ();
      m_pipe = m_scoreBoard.SetPipe (m_txBuffer->HeadSequence (), m_highTxMark, m_highRxt, m_retxThresh);
      if (m_cWnd.Get () / m_segmentSize > m_pipe)
        {
          if (m_cWnd.Get () / m_segmentSize - m_pipe >= 1)
            {
              SendDataPacket (m_highTxMark, m_segmentSize, true);
            }
        }
    }
}

void
TcpSack::Retransmit (void)
{
  NS_LOG_FUNCTION (this);
  NS_LOG_LOGIC (this << " ReTxTimeout Expired at time " << Simulator::Now ().GetSeconds ());

  // If RTO occurs during loss recovery,
  // reset recovery point and terminate the current fast recovery phase
  if (m_inFastRec)
    {
      if (m_afterRTO)
        {
          m_recover = m_highTxMark;
          m_inFastRec = false;
          m_afterRTO = false;
        }
    }

  // If erroneous timeout in closed/timed-wait state, just return
  if (m_state == CLOSED || m_state == TIME_WAIT)
    {
      return;
    }
  // If all data are received (non-closing socket and nothing to send), just return
  if (m_state <= ESTABLISHED && m_txBuffer->HeadSequence () >= m_highTxMark)
    {
      return;
    }

  // According to RFC2581 sec.3.1, upon RTO, ssthresh is set to half of flight
  // size and cwnd is set to 1*MSS, then the lost packet is retransmitted and
  // TCP back to slow start
  m_ssThresh = std::max (2 * m_segmentSize, BytesInFlight () / 2);
  m_cWnd = m_segmentSize;
  m_nextTxSequence = m_txBuffer->HeadSequence (); // Restart from highest Ack
  NS_LOG_INFO ("RTO. Reset cwnd to " << m_cWnd <<
               ", ssthresh to " << m_ssThresh << ", restart from seqnum " << m_nextTxSequence);
  uint32_t sz = DoRetransmitPacket ();                          // Retransmit the packet

  // Update m_highRxt to prevent duplicate retransmission
  m_highRxt = std::max (m_txBuffer->HeadSequence () + sz, m_highRxt.Get ());
}

void
TcpSack::SetSegSize (uint32_t size)
{
  NS_LOG_FUNCTION (this << size);
  NS_ABORT_MSG_UNLESS (m_state == CLOSED, "TcpSack::SetSegSize() cannot change segment size after connection started.");
  m_segmentSize = size;
}

void
TcpSack::SetInitialSSThresh (uint32_t threshold)
{
  NS_LOG_FUNCTION (this << threshold);
  m_ssThresh = threshold;
}

uint32_t
TcpSack::GetInitialSSThresh (void) const
{
  NS_LOG_FUNCTION (this);
  return m_ssThresh;
}

void
TcpSack::SetInitialCwnd (uint32_t cwnd)
{
  NS_LOG_FUNCTION (this << cwnd);
  NS_ABORT_MSG_UNLESS (m_state == CLOSED, "TcpSack::SetInitialCwnd() cannot change initial cwnd after connection started.");
  m_initialCWnd = cwnd;
}

uint32_t
TcpSack::GetInitialCwnd (void) const
{
  NS_LOG_FUNCTION (this);
  return m_initialCWnd;
}

void
TcpSack::InitializeCwnd (void)
{
  NS_LOG_FUNCTION (this);
  /*
   * Initialize congestion window, default to 1 MSS (RFC2001, sec.1) and must
   * not be larger than 2 MSS (RFC2581, sec.3.1). Both m_initiaCWnd and
   * m_segmentSize are set by the attribute system in ns3::TcpSocket.
   */
  m_cWnd = m_initialCWnd * m_segmentSize;
}

void
TcpSack::ScaleSsThresh (uint8_t scaleFactor)
{
  m_ssThresh <<= scaleFactor;
}

uint32_t
TcpSack::DoRetransmitPacket (void)
{
  NS_LOG_FUNCTION (this);
  // Retransmit SYN packet
  if (m_state == SYN_SENT)
    {
      if (m_cnCount > 0)
        {
          SendEmptyPacket (TcpHeader::SYN);
        }
      else
        {
          NotifyConnectionFailed ();
        }
      return 0;
    }
  // Retransmit non-data packet: Only if in FIN_WAIT_1 or CLOSING state
  if (m_txBuffer->Size () == 0)
    {
      if (m_state == FIN_WAIT_1 || m_state == CLOSING)
        { // Must have lost FIN, re-send
          SendEmptyPacket (TcpHeader::FIN);
        }
      return 0;
    }
  // Retransmit a data packet: Call SendDataPacket
  NS_LOG_LOGIC ("TcpSocketBase " << this << " retxing seq " << m_txBuffer->HeadSequence ());
  uint32_t sz = SendDataPacket (m_txBuffer->HeadSequence (), m_segmentSize, true);
  // In case of RTO, advance m_nextTxSequence
  m_nextTxSequence = std::max (m_nextTxSequence.Get (), m_txBuffer->HeadSequence () + sz);
  return sz;
}

void
TcpSack::StartPipe (void)
{
  NS_LOG_FUNCTION (this);

  bool updateHighRxt;
  bool continuePipe = true;
  bool rule2 = false;

  // Find the number of octets currently in transit
  m_pipe = m_scoreBoard.SetPipe (m_txBuffer->HeadSequence (), m_highTxMark, m_highRxt, m_retxThresh);
  NS_LOG_LOGIC ("Number of packets currently in transit: " << m_pipe);

  SequenceNumber32 txSeq = SequenceNumber32 (0);
  uint32_t size = 0;

  // Take advantage of additional available m_cWnd if possible
  NS_LOG_LOGIC ("Congestion window in number of packets: " << m_cWnd.Get () / m_segmentSize);

  if (m_cWnd.Get () / m_segmentSize > m_pipe)
    {
      while ((m_cWnd.Get () / m_segmentSize - m_pipe >= 1) && continuePipe)
        {
          NS_LOG_LOGIC (m_cWnd.Get () / m_segmentSize - m_pipe);

          updateHighRxt = true;

          // Find the next sequence number to transmit
          if (m_scoreBoard.NextSeqPerRule1 (m_highRxt, txSeq, m_retxThresh))
            { // Rule 1 in NexSeq () (RFC 3517)
              size = SendDataPacket (txSeq, m_segmentSize, true);
            }
          else if (SendPendingData (m_connected))
            { // Rule 2 in NextSeq ()
              rule2 = true;
            }
          else if (m_scoreBoard.NextSeqPerRule3 (m_highRxt, txSeq))
            { // Rule 3 in NextSeq ()
              size = SendDataPacket (txSeq, m_segmentSize, true);
            }
          else if (m_txBuffer->HeadSequence () > m_rescueRxt)
            {
              if (m_scoreBoard.NextSeqPerRule4 (m_txBuffer->HeadSequence (), txSeq, m_highTxMark, m_highRxt, m_retxThresh))
                { // Rule 4 in NextSeq ()
                  size = SendDataPacket (txSeq, m_segmentSize, true);
                  m_rescueRxt = m_recover;
                  updateHighRxt = false;
                }
            }

          // Update highTxMark, highRxt, and nextTxSequence
          if (txSeq != SequenceNumber32 (0))
            {
              if ((txSeq < (m_highTxMark - size)) && updateHighRxt)
                {
                  m_highRxt = std::max (txSeq + size, m_highRxt.Get ());
                }
              else
                {
                  m_highTxMark = std::max (txSeq + size, m_highTxMark.Get ());
                  m_nextTxSequence += size;
                }
            }
          NS_LOG_LOGIC ("The transmitted sequence number: " << txSeq);
          NS_LOG_LOGIC ("HighTxMark: " << m_highTxMark << " HighRxt: " << m_highRxt << " NextTxSequence: " << m_nextTxSequence);

          // increment the amount of outstanding data
          m_pipe++;

          if (txSeq == SequenceNumber32 (0) && rule2)
            { // No more data to send
              continuePipe = false;
            }
        }
    }
}

} // namespace ns3
