/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2015 ResiliNets, ITTC, University of Kansas
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Authors: Truc Anh N. Nguyen <annguyen@ittc.ku.edu>
 *
 * James P.G. Sterbenz <jpgs@ittc.ku.edu>, director
 * ResiliNets Research Group  http://wiki.ittc.ku.edu/resilinets
 * Information and Telecommunication Technology Center (ITTC)
 * and Department of Electrical Engineering and Computer Science
 * The University of Kansas Lawrence, KS USA.
 */


#ifndef TCPSACK_H_
#define TCPSACK_H_

#include "tcp-socket-base.h"
#include "tcp-option.h"
#include "tcp-option-sack-permitted.h"
#include "tcp-option-sack.h"
#include "tcp-scoreboard.h"

namespace ns3 {

typedef std::pair<SequenceNumber32, SequenceNumber32> SackBlock;
typedef std::list<SackBlock> SackList;

class Packet;

/**
 * \ingroup socket
 * \ingroup tcp
 *
 * \brief An implementation of a stream socket using TCP.
 *
 * This class contains the implementation of the conservative
 * selective acknowledgment (SACK)-based loss recovery algorithm for TCP
 * as described in \RFC{6675}.
 *
 * SACK blocks are generated based on \RFC{2018}
 */
class TcpSack : public TcpSocketBase
{
public:
  static TypeId GetTypeId (void);

  TcpSack (void);
  TcpSack (const TcpSack& sock);
  virtual ~TcpSack (void);

  // From TcpSocketBase
  virtual int Connect (const Address &address);
  virtual int Listen (void);

protected:
  /**
   * \brief Limit the size of outstanding data based on the cwnd and the receiver's advertised window
   *
   * \return the max. possible number of unacked bytes
   */
  virtual uint32_t Window (void);

  /**
   * \brief Call CopyObject<TcpWestwood> to clone me
   */
  virtual Ptr<TcpSocketBase> Fork (void);

  /**
   * \brief Adjust the cwnd based on the current congestion control phase
   * Run the "pipe" algorithm
   *
   * \param seq The acknowledgment number
   */
  virtual void NewAck (SequenceNumber32 const& seq);

  /**
   * \brief This method is not implemented in TCP SACK
   *        Its functionality is replaced by the private method SackDupAck()
   */
  virtual void DupAck (const TcpHeader& header, uint32_t count);

  /**
   * \brief Received an ACK packet
   * \param packet the packet
   * \param tcpHeader the packet's TCP header
   */
  virtual void ReceivedAck (Ptr<Packet> packet, const TcpHeader& tcpHeader);

  /**
   * \brief Retransmit upon an RTO event
   */
  virtual void Retransmit (void);

  /**
   * \brief Retransmit a packet starting with first sequence in tx buffer
   */
  uint32_t DoRetransmitPacket (void);

  // Implementing ns3::TcpSocket -- Attribute get/set
  /**
   * \param size the segment size to be used in a connection
   */
  virtual void SetSegSize (uint32_t size);

  /**
   * \param threshold the slow-start threshold
   */
  virtual void SetInitialSSThresh (uint32_t threshold);

  /**
   * \return the slow-start threshold
   */
  virtual uint32_t GetInitialSSThresh (void) const;

  /**
   * \param cwnd the initial cwnd
   */
  virtual void SetInitialCwnd (uint32_t cwnd);

  /**
   * \return the initial cwnd
   */
  virtual uint32_t GetInitialCwnd (void) const;
  virtual void ScaleSsThresh (uint8_t scaleFactor);

private:
  /**
   * \brief Initialize cwnd at the beginning of a connection
   */
  void InitializeCwnd (void);

  /**
   * \brief Calculate the number of bytes currently in transit
   *        and take advantage of available m_cWnd to send more data if possible
   */
  void StartPipe ();

  /**
   * \brief Retransmit the missing packet, and enter fast recovery if m_retxThresh DupAcks are received
   *        Run the "pipe" algorithm to perform additional transmissions if possible
   *
   * \param header The TCP header of the ACK packet
   * \param count The number of TCP SACK DupAcks received
   */
  void SackDupAck (const TcpHeader& header, uint32_t count);

protected:
  TracedValue<uint32_t>  m_cWnd;         //!< Congestion window
  uint32_t               m_ssThresh;     //!< Slow Start Threshold
  uint32_t               m_initialCWnd;  //!< Initial cWnd value
  SequenceNumber32       m_recover;      //!< Previous highest Tx seqnum for fast recovery
  uint32_t               m_retxThresh;   //!< Fast Retransmit threshold
  bool                   m_inFastRec;    //!< Currently in fast recovery if TRUE
  bool                   m_limitedTx;    //!< Perform limited transmit

  TracedValue<SequenceNumber32>      m_highRxt;          //!< Highest sequence number which has been retransmitted
  TracedValue<SequenceNumber32>      m_rescueRxt;        //!< Highest sequence number which has been optimistically retransmitted
  uint32_t                           m_pipe;             //!< Number of outstanding bytes
  uint32_t                           m_sackDupAckCount;  //!< TCP SACK dupack (as defined in \RFC{6675}) counter



};

} // namespace ns3

#endif /* TCPSACK_H_ */
