/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2015 ResiliNets, ITTC, University of Kansas
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Authors: Truc Anh N. Nguyen <annguyen@ittc.ku.edu>
 *
 * James P.G. Sterbenz <jpgs@ittc.ku.edu>, director
 * ResiliNets Research Group  http://wiki.ittc.ku.edu/resilinets
 * Information and Telecommunication Technology Center (ITTC)
 * and Department of Electrical Engineering and Computer Science
 * The University of Kansas Lawrence, KS USA.
 */

#include <iostream>
#include <algorithm>
#include <cstring>
#include "ns3/log.h"
#include "ns3/uinteger.h"
#include "ns3/trace-source-accessor.h"

#include "tcp-scoreboard.h"

namespace ns3 {

NS_LOG_COMPONENT_DEFINE ("TcpScoreBoard");

NS_OBJECT_ENSURE_REGISTERED (TcpScoreBoard);

TypeId
TcpScoreBoard::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::TcpScoreBoard")
    .SetParent<Object> ()
    .AddConstructor<TcpScoreBoard> ()
    .AddAttribute ("MaxScoreBoardSize",
                   "Max size of the scoreboard",
                   UintegerValue (20000000),
                   MakeUintegerAccessor (&TcpScoreBoard::m_maxScoreBoardSize),
                   MakeUintegerChecker<uint32_t> ())
    .AddTraceSource ("HighestSACKedSequence",
                     "Highest sequence number that has been SACKed",
                     MakeTraceSourceAccessor (&TcpScoreBoard::m_highSack),
                     "ns3::SequenceNumber32TracedValueCallback")

  ;
  return tid;
}

TcpScoreBoard::TcpScoreBoard (void)
  : m_maxScoreBoardSize (20000000),
    m_scoreBoard (0),
    m_highSack (0)

{
}

TcpScoreBoard::~TcpScoreBoard (void)
{
}

void
TcpScoreBoard::SetMaxScoreBoardSize (uint32_t maxSize)
{
  NS_LOG_FUNCTION (this << maxSize);
  m_maxScoreBoardSize = maxSize;
}

uint32_t
TcpScoreBoard::GetMaxScoreBoardSize (void) const
{
  NS_LOG_FUNCTION (this);
  return m_maxScoreBoardSize;
}

SequenceNumber32
TcpScoreBoard::GetHighSack (void) const
{
  NS_LOG_FUNCTION (this);
  return m_highSack;
}

bool
TcpScoreBoard::AddEntry (TcpScoreBoardEntry& entry)
{
  NS_LOG_FUNCTION (this);
  NS_LOG_LOGIC ("Scoreboard size: " << m_scoreBoard.size ());
  NS_LOG_LOGIC ("Max Scoreboard size: " << m_maxScoreBoardSize);
  if (m_scoreBoard.size () < m_maxScoreBoardSize) // Add entry if space available
    {
      if (!m_scoreBoard.empty ())
        { // Make sure the entry is a new entry
          for (BufIterator i = m_scoreBoard.begin (); i != m_scoreBoard.end (); ++i)
            {
              if ((i)->GetSequenceNumber () == entry.GetSequenceNumber ())
                {
                  return false;
                }
            }
        }
      m_scoreBoard.push_back (entry);
      NS_LOG_LOGIC ("Packet's sequence number " << entry.GetSequenceNumber () << " added to the scoreboard.");
      return true;
    }
  NS_LOG_LOGIC ("Rejected. Scoreboard is full.");
  return false;
}

bool
TcpScoreBoard::IsDupAck (std::list<std::pair<SequenceNumber32, SequenceNumber32> >& sackList, const SequenceNumber32& highAck, const SequenceNumber32& highData)
{
  NS_LOG_FUNCTION (this << highAck << highData);
  NS_LOG_LOGIC ("m_scoreBoard.size ()" << m_scoreBoard.size ());
  if (m_scoreBoard.empty ())
    {
      return false;
    }
  for (SackIterator j = sackList.begin (); j != sackList.end (); ++j)
    {
      SequenceNumber32 leftEdge = (*j).first;
      SequenceNumber32 rightEdge = (*j).second;
      NS_LOG_LOGIC ("left edge: " << leftEdge << " right edge: " << rightEdge);
      for (BufIterator i = m_scoreBoard.begin (); i != m_scoreBoard.end (); ++i)
        {
          SequenceNumber32 seqNum = i->GetSequenceNumber ();
          bool isSacked = i->GetSackFlag ();
          if (seqNum >= leftEdge && seqNum < rightEdge && highAck <= seqNum && seqNum < highData && !isSacked)
            { // The block identifies the octets that are unACKed, un-SACKed and between highAck and highData
              return true;
            }
        }
    }
  return false;
}

void
TcpScoreBoard::Update (const SequenceNumber32& ackNumber, std::list<std::pair<SequenceNumber32, SequenceNumber32> >& sackList)
{
  NS_LOG_FUNCTION (this);

  // No entry in the scoreboard, return
  if (m_scoreBoard.empty ())
    {
      return;
    }

  // Scan the scoreboard to remove data that are cummulatively ACKed
  BufIterator i = m_scoreBoard.begin ();
  while (i != m_scoreBoard.end ())
    {
      if (i->GetSequenceNumber () < ackNumber)
        {
          BufIterator j = i;
          ++i;
          m_scoreBoard.erase (j);
          if (m_scoreBoard.empty ())
            {
              break;
            }
        }
      else
        {
          break;
        }
    }

  // Scan the scoreboard to update the SACK flag
  if (!sackList.empty ())
    {
      SequenceNumber32 sackLeft;   // left edge of a SACK block
      SequenceNumber32 sackRight;  // right edge of a SACK block
      for (SackIterator j = sackList.begin (); j != sackList.end (); ++j)
        {
          sackLeft = (*j).first;
          sackRight = (*j).second;
          UpdateSackFlag (sackLeft, sackRight);
        }
    }

}

void
TcpScoreBoard::UpdateSackFlag (const SequenceNumber32& leftEdge, const SequenceNumber32& rightEdge)
{
  NS_LOG_FUNCTION (this << leftEdge << rightEdge);

  // No entry in the scoreboard, return
  if (m_scoreBoard.empty ())
    {
      return;
    }

  for (BufIterator i = m_scoreBoard.begin (); i != m_scoreBoard.end (); ++i)
    {
      if (i->GetSequenceNumber () >= leftEdge && i->GetSequenceNumber () < rightEdge)
        {
          // Set SACK flag if it hasn't been set
          if (!(i->GetSackFlag ()))
            {
              i->SetSackFlag (true);
              m_highSack = std::max (rightEdge, m_highSack.Get ());
              NS_LOG_LOGIC ("Highest SACKed sequence number: " << m_highSack);
            }
        }
    }
}

bool
TcpScoreBoard::IsLost (const SequenceNumber32& seq, uint32_t dupThresh)
{
  NS_LOG_FUNCTION (this << seq << dupThresh);

  // No entry in the scoreboard, return
  if (m_scoreBoard.empty ())
    {
      return false;
    }

  uint32_t count = 0;    // Count the number of SACKed packets having arrived above seq

  for (BufIterator i = m_scoreBoard.begin (); i != m_scoreBoard.end (); ++i)
    {
      if (i->GetSequenceNumber () <= seq)
        {
          continue;
        }
      else
        {
          if (i->GetSackFlag ())
            {
              count++;
            }
        }
    }

  if (count >= dupThresh)
    {
      return true;    // Return True if dupThresh SACKed packets arrived above seq
    }
  else
    {
      return false;
    }
}

SequenceNumber32
TcpScoreBoard::FindNextRetx (SequenceNumber32 startSeq)
{
  NS_LOG_FUNCTION (this << startSeq);

  // No entry in the scoreboard, return
  if (m_scoreBoard.empty ())
    {
      return SequenceNumber32 (0);
    }

  for (BufIterator i = m_scoreBoard.begin (); i != m_scoreBoard.end (); ++i)
    {
      SequenceNumber32 rtxSeq = i->GetSequenceNumber ();
      NS_LOG_LOGIC ("rtxSeq: " << rtxSeq << " startSeq: " << startSeq);
      if (rtxSeq < startSeq)
        {
          continue;
        }
      else
        {
          // the next sequence to be retransmitted satisfies the following conditions (\RFC{2018}):
          // (1) m_highSack > rtxSeq >= startSeq
          // (2) rtxSeq has not been SACKed
          if (!(i->GetSackFlag ()) && rtxSeq < m_highSack)
            {
              return rtxSeq;
            }
        }
    }
  return SequenceNumber32 (0);

}

uint32_t
TcpScoreBoard::SetPipe (const SequenceNumber32& highAck, const SequenceNumber32& highData, const SequenceNumber32& highRxt, uint32_t dupThresh)
{
  NS_LOG_FUNCTION (this << highAck << highData << highRxt << dupThresh);

  // No entry in the scoreboard, return
  if (m_scoreBoard.empty ())
    {
      return 0;
    }

  uint32_t pipe = 0;   // Keep track of the number of outstanding packets

  for (BufIterator i = m_scoreBoard.begin (); i != m_scoreBoard.end (); ++i)
    {
      SequenceNumber32 seqNum = i->GetSequenceNumber ();
      bool isSacked = i->GetSackFlag ();
      if (highAck <= seqNum && seqNum < highData && !isSacked)
        { // Only look at packet whose sequence number between highAck and highData that has not been SACKed
          if (seqNum < highRxt || !(IsLost (seqNum, dupThresh)))
            { // Outstanding packets are packets that satisfy one of the following conditions:
              // (1) have not been SACKed and have not been determined to be lost
              // (2) have been retransmitted
              pipe++;
            }
        }
    }
  return pipe;
}

bool
TcpScoreBoard::NextSeqPerRule1 (const SequenceNumber32& highRxt, SequenceNumber32& nextSeq, uint32_t dupThresh)
{
  NS_LOG_FUNCTION (this << highRxt << dupThresh);

  if (m_scoreBoard.empty ())
    {
      return false;
    }

  for (BufIterator i = m_scoreBoard.begin (); i != m_scoreBoard.end (); ++i)
    {
      SequenceNumber32 seq = i->GetSequenceNumber ();
      bool isSacked = i->GetSackFlag ();
      if (isSacked)
        {
          continue;
        }
      else
        {
          if (seq >= highRxt && seq < m_highSack && IsLost (seq,dupThresh))
            { // Rule 1.a: seq is greater than HighRxt
              // Rule 1.b: seq is less than the highest SACKed sequence number
              // Rule 1.c: seq is determined to be lost
              nextSeq = seq;
              return true;
            }
        }
    }
  return false;
}

bool
TcpScoreBoard::NextSeqPerRule3 (const SequenceNumber32& highRxt, SequenceNumber32& nextSeq)
{
  NS_LOG_FUNCTION (this << highRxt);

  if (m_scoreBoard.empty ())
    {
      return false;
    }

  for (BufIterator i = m_scoreBoard.begin (); i != m_scoreBoard.end (); ++i)
    {
      SequenceNumber32 seq = i->GetSequenceNumber ();
      bool isSacked = i->GetSackFlag ();
      if (isSacked)
        {
          continue;
        }
      else
        {
          if (seq >= highRxt && seq < m_highSack)
            { // Rule 1.a: seq is greater than HighRxt
              // Rule 1.b: seq is less than the highest SACKed sequence number
              nextSeq = seq;
              return true;
            }
        }
    }
  return false;
}

bool
TcpScoreBoard::NextSeqPerRule4 (const SequenceNumber32& highAck, SequenceNumber32& nextSeq, const SequenceNumber32& highData, const SequenceNumber32& highRxt, uint32_t dupThresh)
{
  NS_LOG_FUNCTION (this);
  nextSeq = SequenceNumber32 (0);

  // Find the highest outstanding unSACKed sequence number
  if (m_scoreBoard.empty ())
    {
      return false;
    }
  else
    {
      for (BufIterator i = m_scoreBoard.begin (); i != m_scoreBoard.end (); ++i)
        {
          SequenceNumber32 seqNum = i->GetSequenceNumber ();
          bool isSacked = i->GetSackFlag ();
          if (highAck <= seqNum && seqNum < highData && !isSacked)
            {   // Only look at packet whose sequence number between highAck and highData that has not been SACKed
              if (seqNum < highRxt || !(IsLost (seqNum, dupThresh)))
                {   // Outstanding packets are packets that satisfy one of the following conditions:
                    // (1) have not been SACKed and have not been determined to be lost
                    // (2) have been retransmitted
                  if (seqNum > nextSeq)
                    {
                      nextSeq = seqNum;
                    }
                }
            }
        }
    }

  if (nextSeq != SequenceNumber32 (0))
    {
      return true;
    }
  else
    {
      return false;
    }
}

void
TcpScoreBoard::Reset ()
{
  NS_LOG_FUNCTION (this);
  for (BufIterator i = m_scoreBoard.begin (); i != m_scoreBoard.end (); ++i)
    {
      bool isSacked = i->GetSackFlag ();
      if (isSacked)
        {
          i->SetSackFlag (false);
        }
    }

  m_highSack = 0;
}


//
// TcpScoreBoardEntry
//

NS_OBJECT_ENSURE_REGISTERED (TcpScoreBoardEntry);

TypeId
TcpScoreBoardEntry::GetTypeId (void)
{
  static TypeId tid = TypeId ("ns3::TcpScoreBoardEntry")
    .SetParent<Object> ()
    .AddConstructor<TcpScoreBoardEntry> ()

  ;
  return tid;
}

TcpScoreBoardEntry::TcpScoreBoardEntry (SequenceNumber32 seq, bool sacked)
  : m_sequenceNumber (seq),
    m_isSacked (sacked)
{
}

TcpScoreBoardEntry::~TcpScoreBoardEntry (void)
{
}

void
TcpScoreBoardEntry::SetSequenceNumber (SequenceNumber32 seq)
{
  m_sequenceNumber = seq;
}

SequenceNumber32
TcpScoreBoardEntry::GetSequenceNumber (void) const
{
  return m_sequenceNumber;
}

void
TcpScoreBoardEntry::SetSackFlag (bool isSacked)
{
  m_isSacked = isSacked;
}

bool
TcpScoreBoardEntry::GetSackFlag (void) const
{
  return m_isSacked;
}

}

