/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 * Copyright (c) 2015 ResiliNets, ITTC, University of Kansas
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Authors: Truc Anh N. Nguyen <annguyen@ittc.ku.edu>
 *
 * James P.G. Sterbenz <jpgs@ittc.ku.edu>, director
 * ResiliNets Research Group  http://wiki.ittc.ku.edu/resilinets
 * Information and Telecommunication Technology Center (ITTC)
 * and Department of Electrical Engineering and Computer Science
 * The University of Kansas Lawrence, KS USA.
 */

#ifndef TCPSCOREBOARD_H_
#define TCPSCOREBOARD_H_

#include <list>

#include "ns3/object.h"
#include "ns3/ptr.h"
#include "ns3/sequence-number.h"
#include "ns3/traced-value.h"

namespace ns3 {

class Packet;

/**
 * \ingroup tcp
 *
 * \brief These classes contain the implementation of a "scoreboard".
 * The scoreboard is used by a TCP sender to keep track of SACK information.
 * The implementation is based on the specification in \RFC{6675}
 *
 */

/**
 * \brief The TcpScoreBoardEntry class implements an entry in the scoreboard.
 * Each entry has 2 fields.
 * The first field stores the transmitted segment's sequence number;
 * the second is the SACK flag field.
 */

class TcpScoreBoardEntry : public Object
{
public:
  static TypeId GetTypeId (void);

  TcpScoreBoardEntry (SequenceNumber32 seq = SequenceNumber32(0), bool isSacked = false);

  virtual ~TcpScoreBoardEntry (void);

  /**
   * \param seq the packet's sequence number
   */
  void SetSequenceNumber (SequenceNumber32 seq);

  /*
   * \return the packet's sequence number
   */
  SequenceNumber32 GetSequenceNumber (void) const;

  /**
   * \param isSacked the packet's SACK flag
   */
  void SetSackFlag (bool isSacked);

  /**
   * \return the packet's SACK flag
   */
  bool GetSackFlag (void) const;

private:
  SequenceNumber32 m_sequenceNumber;                //!< Packet's sequence number
  bool m_isSacked;                                  //!< True if the packet is SACKed

};

/**
 * \brief This class implements the scoreboard,
 * which is a list of TcpScoreBoardEntry entries.
 * Each entry is specified by TcpScoreBoardEntry
 */
class TcpScoreBoard : public Object
{
public:
  static TypeId GetTypeId (void);
  TcpScoreBoard ();
  virtual ~TcpScoreBoard (void);

  /**
   * \brief Set the maximum number of entries the scoreboard can hold
   *
   * \param maxSize the maximum number of entries the scoreboard can hold
   */
  void SetMaxScoreBoardSize (uint32_t maxSize);

  /**
   * \brief Get the maximum capacity of the scoreboard
   *
   * \return the maximum scoreboard size
   */
  uint32_t GetMaxScoreBoardSize (void) const;

  /**
   * \param Get the highest sequence number that has been SACKed
   *
   * \return the highest sequence number having been SACKed
   */
  SequenceNumber32 GetHighSack (void) const;

  /**
   * \brief Add an entry to the scoreboard if space available
   *
   * \param entry the entry to be added to the scoreboard
   * \return true if success
   */
  bool AddEntry (TcpScoreBoardEntry& entry);

  /**
   * \brief Determine if the segment carries the SACK list is a duplicate ACK as defined in \RFC{6675}
   *
   * \param sackList the list of all SACK blocks received in this segment
   * \param highAck the highest sequence number that has been cumulatively ACKed
   * \param highData the highest sequence number that has been transmitted
   * \return true if the segment is a duplicate ACK; false otherwise
   */
  bool IsDupAck (std::list<std::pair<SequenceNumber32, SequenceNumber32> >& sackList, const SequenceNumber32& highAck, const SequenceNumber32& highData);

  /**
   * \brief Update the scoreboard upon the receipt of an ACK
   *
   * \param ackNumber the recently received acknowledgment number
   * \param sackList the list of all SACK blocks received in this ACK
   */
  void Update (const SequenceNumber32& ackNumber, std::list<std::pair<SequenceNumber32, SequenceNumber32> >& sackList);

  /**
   * \brief Check to see if a given packet is considered to be lost as specified in \RFC{6675}
   *
   * \param seq the sequence number of the packet being examined
   * \param dupThresh the duplicate ACK threshold
   * \return true if the packet is considered lost
   */
  bool IsLost (const SequenceNumber32& seq, uint32_t dupThresh);

  /**
   * \brief Calculate the number of packets currently in transit
   *
   * \param highAck the highest byte of data that has been cumulatively ACKed
   * \pram highData the highest sequence number transmitted
   * \param highRxt the highest sequence number that has been retransmitted
   * \param dupThresh the duplicate ACK threshold
   * \return the number of outstanding packets
   */
  uint32_t SetPipe (const SequenceNumber32& highAck, const SequenceNumber32& highData, const SequenceNumber32& highRxt, uint32_t dupThresh);

  /**
   * \brief Determine the next sequence number to be transmitted based on Rule 1 in \RFC{6675}
   *
   * \param highRxt the highest sequence number that has been retransmitted
   * \param nextSeq the next sequence number to be transmitted
   * \return true if there exists a sequence number satisfying Rule 1; false otherwise
   */
  bool NextSeqPerRule1 (const SequenceNumber32& highRxt, SequenceNumber32& nextSeq, uint32_t dupThresh);

  /**
   * \brief Determine the next sequence number to be transmitted based on Rule 3 in \RFC{6675}
   *
   * \param highRxt the highest sequence number that has been retransmitted
   * \param nextSeq the next sequence number to be transmitted
   * \return true if there exits a sequence number satisfying Rule 3; false otherwise
   */
  bool NextSeqPerRule3 (const SequenceNumber32& highRxt, SequenceNumber32& nextSeq);

  /**
   * \brief Determine the next sequence number to be transmitted based on Rule 4 in \RFC{6675}
   *
   * \param highAck the highest byte of data that has been cumulatively ACKed
   * \param nextSeq the next sequence number to be transmitted
   * \param highData the highest sequence number transmitted
   * \param highRxt the highest sequence number that has been retransmitted
   * \param dupThresh the duplicate ACK threshold
   * \return true if there exists a sequence number satisfying rule 4; false otherwise
   */
  bool NextSeqPerRule4 (const SequenceNumber32& highAck, SequenceNumber32& nextSeq, const SequenceNumber32& highData, const SequenceNumber32& highRxt, uint32_t dupThresh);

  /**
   * \brief Find the next segment for retransmission based on \RFC{2018}, i.e.
   * segment with SACKed bit off & < highest SACKed segment
   * \param startSeq the sequence number to start searching
   * \return the sequence number to be retransmitted
   */
  SequenceNumber32 FindNextRetx (SequenceNumber32 startSeq);

  /**
   * Reset the SACKed flags
   */
  void Reset ();

private:
  /**
   * \brief Update SACK flag of the corresponding entries in the scoreboard
   *
   *  based on the leftEdge and rightEdge of a SACK block
   * \param leftEdge the left edge of the block
   * \param rightEdge the right edge of the block
   */
  void UpdateSackFlag (const SequenceNumber32& leftEdge, const SequenceNumber32& rightEdge);

private:
  typedef std::list<TcpScoreBoardEntry>::iterator BufIterator;
  typedef std::list<std::pair<SequenceNumber32, SequenceNumber32> >::iterator SackIterator;

  uint32_t m_maxScoreBoardSize;                                   //!< Max number of data bytes in the scoreboard
  std::list<TcpScoreBoardEntry> m_scoreBoard;                     //!< Corresponding data (may be null)
  TracedValue<SequenceNumber32> m_highSack;                       //!< Highest sequence number having been SACKed

};
}

#endif /* TCPSCOREBOARD_H_ */
